(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)




(*** types ***)


type config = Config.config
type choice_point = State.choice_point
type term = Term.term
type literal = Term.literal
type subst = Subst.subst

type complexity = int



type bound = {
  config: config;

  (* the bound to obey *)
  mutable bound: complexity;

  (* how many incomplete exhausted branches have been found and backtracked from? *)
  mutable incomplete_branches: int;

  (* how many restarts with an increased depth bound have been done? *)
  mutable restarts: int;

  (* is the current derivation incomplete,
     i.e. an incomplete exhausted branch has been found, but was ignored? *)
  mutable derivation_incomplete: bool;

  (* lowest complexity of among the dropped candidates. *)
  mutable dropped_complexity: complexity option;

  (* oldest choice point among the dropped candidates. *)
  mutable dropped_choice_point: choice_point option;
}





(*** creation ***)

let create (config: config) : bound =
  {
    config = config;
    bound = Config.deepening_bound config;
    
    incomplete_branches = 0;
    restarts = 0;
    
    derivation_incomplete = false;
    dropped_choice_point = None;
    dropped_complexity = None;
  }

let restart ?(keep_bound: bool = false) (bound: bound) : unit =
  if not keep_bound then begin
    match bound.dropped_complexity with
      | None ->
	  failwith "Bound.restart"
	    
      | Some dropped_complexity ->
	  bound.bound <- dropped_complexity
  end;

  bound.restarts <- bound.restarts + 1;
  bound.derivation_incomplete <- false;
  bound.dropped_complexity <- None;
  bound.dropped_choice_point <- None



(*** representation ***)


let complexity_to_string (complexity: complexity) : string =
  string_of_int complexity


let print_statistic (bound: bound) : unit =
  Print.print_statistic "Incomplete Branches" bound.incomplete_branches;
  Print.print_statistic "Restarts" bound.restarts;
  Print.print_statistic "Bound" bound.bound




(*** access ***)

let current_bound (bound: bound) : complexity =
  bound.bound
    
let dropped_complexity (bound: bound) : complexity option =
  bound.dropped_complexity
  
let dropped_choice_point (bound: bound) : choice_point option =
  bound.dropped_choice_point




(*** complexity ***)



let get_complexity (bound: bound) (literal: literal) : complexity =
  if Config.iterative_deepening bound.config = Flags.IT_TermDepth then
    Term_attributes.depth_of_literal literal
  else
    Term_attributes.weight_of_literal literal

let get_complexity_subst (bound: bound) (literal: literal) (subst: subst) (offset: int) : complexity =
  if Config.iterative_deepening bound.config = Flags.IT_TermDepth then
    Term_attributes.depth_of_literal_subst literal subst offset
  else
    Term_attributes.weight_of_literal_subst literal subst offset
      
let compare_complexity (complexity1: complexity) (complexity2: complexity) : int =
  Tools.compare_int complexity1 complexity2





(*** check ***)

(* convention: methods ending with _
   take a literal's complexity instead of the literal itself *)

(* does the literal exceed the deepening bound? *)
let exceeds (bound: bound) (complexity: complexity) : bool =
  compare_complexity bound.bound complexity = -1


(* does this exceeding candidate have the lowest complexity
   of all up to now dropped candidates? *)
let exceeds_current_bound_ (bound: bound) (complexity: complexity) : bool =
  match bound.dropped_complexity with
    | None ->
	(* first exceeding candidate *)
	true
	  
    | Some dropped_complexity ->
	(* lowest candidate? *)
	compare_complexity dropped_complexity complexity = 1

let exceeds_current_bound (bound: bound) (literal: literal) : bool =
  exceeds_current_bound_ bound (get_complexity bound literal)


(* for backtracking over an incomplete branch:
   is the choice point of this candidate of any relevance? *)
let exceeds_current_choice_point (bound: bound) (choice_point: choice_point) : bool =
  match bound.dropped_choice_point with
    | None ->
	true
	  
    | Some dropped_choice_point ->
	not (State.backtracking_depends_on choice_point dropped_choice_point)


let exceeds_current (bound: bound) (complexity: complexity) (choice_point: choice_point) : bool =
  exceeds bound complexity
  &&
    (
      exceeds_current_bound_ bound complexity
      ||
      exceeds_current_choice_point bound choice_point
    )


(* the literal is dropped if it exceeds the bound
   and the complexity and choice point are registered for restarting. *)
(*  method private virtual do_register: complexity -> choice_point -> bool*)
let register' (bound: bound) (complexity: complexity) (choice_point: choice_point) : bool =
  if exceeds bound complexity then begin
    (* update the minimal dropped complexity. *)
    if exceeds_current_bound_ bound complexity then begin
      bound.dropped_complexity <- Some complexity
    end;
    
    (* update the dropped choice point. *)
    if exceeds_current_choice_point bound choice_point then begin
      bound.dropped_choice_point <- Some choice_point;
    end;
    
    true
  end
    
  else begin
    false
  end


let register (bound: bound) (complexity: complexity) (choice_point: choice_point) : bool =
  register' bound complexity choice_point
    
let register_subst (bound: bound) (literal: literal) (subst: subst) (offset: int) (choice_point: choice_point) : bool =
  let complexity =
    get_complexity_subst bound literal subst offset
  in
    register' bound complexity choice_point




(*** incompleteness ***)

let is_derivation_incomplete (bound: bound) : bool =
  bound.derivation_incomplete


let incomplete (bound: bound) : unit =
  bound.derivation_incomplete <- true;
  bound.incomplete_branches <- bound.incomplete_branches + 1




(*** backtracking ***)

let backtrack (bound: bound) : unit =
  match bound.dropped_choice_point with
    | Some choice_point when
	State.is_choice_point_invalid choice_point ->
	bound.dropped_choice_point <- None
	  
    | _ ->
	()
