(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2005
              The University of Iowa

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(* regression algorithm

   the regression alogrithm computes
   - a set of constraints
   - a set of regressed literals

   a constraint is a pairing of two literals with their offsets
   that has to be unified, e.g. (f(x), 0), (f(a), 1).
   after the regression all constraints are unified to get one substitution,
   e.g. the above constraint yields [ 0:x -> 1:a ].

   the regressed literals are the leaves of the regression tree,
   i.e. pairs of literals and offsets, e.g. (g(x, b), 0).

   the regression (conflict set) consists of the application
   of the constraint unifier to the regressed literals,
   in this example { g(a, b) }.


   the actual regression is started from the closing clause,
   which is put paired with a fresh offset into the to_regress list.
   the fresh offset in effect creates a fresh variant of each clause
   used in a context unifier and the regression.
   the constraints and regressed literals are successively built
   buy regressing each clause in the to_regress list,
   which in turn might add new clauses to the regress list.

   a single clause is regressed by regressing each clause literal,
   or more precisely the context literal paired with the clause literal
   in the corresponding context unifier. That is, for the closing clause
   the context literals of the closing context unifier,
   and then in turn the context unifiers of the regressed context literals.
   
   now, there are several possibilities wrt. to regressing a clause literal L
   resp. context literal K:
   
   - Split
   if the context literal K was created by a (left or right) split decision,
   it is not regressed, as it is not implied, but simply guessed.
   and the clause literal L and the clause offset are taken as the regressed literal.
   
   - Assert
   if the context literal K was created by an assert,
   the context unifier creating K is added to the to_regress list with a fresh offset.
   Furthermore, a new constraint between the clause literal L
   and the clause literal L' which was instantiated to K in K's context unifier,
   i.e. the clause literal that was paired with no clause literal in the assert context unifier,
   is added.
   L is paired with its clause context unifier, L' with its fresh offset.

   Thus, the assert literal is replaced by the context literals used to compute it,
   and its instantiation effect is caught in the new constraint.

   - Root Choice Point, -v or +v
   These can not be regressed and are treated as Splits


   Furthermore, there are some limits which may stop the regression early.
   Most importantly, only the assert context literals of the decision level
   in which the conflict occurred are regresed,
   others are also simply treated as Splits.

*)




(*** types ***)



type state = State.state
type choice_point = State.choice_point
type var = Var.var
type literal = Term.literal
type clause = Term.clause
type literal_info = State.literal_info

module LiteralTable = Term.LiteralTable


(* represents an asserted context literal to regress. *)
type to_regress = {
  (* the context literal to regress. *)
  tr_context_literal: literal;

  (* the clause literal paired with tr_context_literal. *)
  tr_clause_literal: literal;

  (* the offset of tr_clause_literal. *)
  tr_clause_offset: int;

  (* the offset of the clause to regress,
     i.e. the clause used in the context unifier computing tr_context_literal. *)
  tr_regress_offset: int;

  (* the information associated with tr_context_literal.
     except for the closing clause this can be computed from tr_context_literal. *)
  tr_regress_info: literal_info;
}


(* the context literals to regress are kept sorted.
   this way, the uip can be detected fairly easily,
   exactly when the smallest and largest context literal to regress are identical.

   as a set only works with a total order variants of the same context literal
   have to be ordered as well, although this is not necessary for the uip.
 *)
module To_regress =
  Set.Make (
    struct
      type t = to_regress

      let compare (x: to_regress) (y: to_regress) : int =
	if x == y then
	  0
	else
	  let cmp =
	    Term.compare_literals x.tr_context_literal y.tr_context_literal
	  in
	  if cmp <> 0 then
	    cmp
          else begin
            (* ensure total order by discriminating the same context literal to regress
	       by its offset, i.e. by variants *)
            if x.tr_regress_offset < y.tr_regress_offset then
              -1
            else if x.tr_regress_offset > y.tr_regress_offset then
              1
            else
              failwith "To_regress.compare"
          end
	    
    end
  )





(* raised to stop the regression early *)
exception MAX_REGRESSED



(* for local debug prints *)
let trace = false





(*** regression ***)



(* if we reached the UIP, stop right here *)
let uip_reached (to_regress: To_regress.t) : bool =
  let first =
    To_regress.max_elt to_regress
  and last =
    To_regress.min_elt to_regress
  in
    Const.first_uip
    &&
    (* ignore the closing clause. *)
    (not (Term.literal_equal first.tr_clause_literal Term.null_literal))
    &&
    (* only one context literal left to regress. *)
    (Term.literal_equal first.tr_context_literal last.tr_context_literal)


(* the uip has been reached,
   so do not regress the remaining context literals,
   but instead just keep the corresponding clause literals. *)
let stop_regression
    (regressed: (literal * int) list)
    (constraints: ((literal * int) * (literal * int)) list)
    (to_regress: To_regress.t)
    : (literal * int) list * ((literal * int) * (literal * int)) list =

  let regressed =
    To_regress.fold
      (fun to_regress regressed ->
	 let new_regressed =
	   (to_regress.tr_clause_literal, to_regress.tr_clause_offset)
	 in
	   new_regressed :: regressed
      )
      to_regress
      regressed
  in
    regressed, constraints







(* regress one context literal *)
let regress_one
    (state: state)
    (* only regress context literals up to this choice point *)
    (regress_stop: choice_point)
    (* clause literal * offset contained in the regressed literal set *)
    (regressed: (literal * int) list)
    (* size of regressed list, passed along for efficiency *)
    (regressed_size: int)
    (* branch nodes to regress * offset *)
    (to_regress: To_regress.t)
    (* fresh offsets for the regression *)
    (next_offset: int ref)

    (* the actual context literal to regress *)
    (current_to_regress: to_regress)
    :
    (* the extended regresset set *)
    ((literal * int) list)
    *
    (* the size of the regressed set *)
    int
    *
    (* the extended literals to regress  *)
    To_regress.t
    =

  (* check for each context literal in the context unifier to regress how to regress it. *)
  let rec regress_one'
      (index: int) (regressed: (literal * int) list) (regressed_size: int) (to_regress: To_regress.t) :
      ((literal * int) list) * int * To_regress.t
      =

    (* all context literals done *)
    if index >= Array.length current_to_regress.tr_regress_info.State.li_context_literals then begin
      regressed, regressed_size, to_regress
    end
    
    (* the assert literal gap, paired with nothing - just ignore it *)
    else if index == current_to_regress.tr_regress_info.State.li_clause_index then begin
      if Const.debug
	&&
	not (Term.literal_equal current_to_regress.tr_regress_info.State.li_context_literals.(index) Term.assert_literal)
      then begin
	failwith "Lemma.regress_one: assert";
      end;
      
      regress_one' (index + 1) regressed regressed_size to_regress
    end

    (* process this context_literal *)
    else begin
      let clause_literal =
	List.nth current_to_regress.tr_regress_info.State.li_clause index
      in
      let context_literal =
	current_to_regress.tr_regress_info.State.li_context_literals.(index)
      in
      let offset =
	current_to_regress.tr_regress_offset
      in

      	(* check type of context literal *)

	(* the default pseudo-literal -v - no regression, just add the clause literal *)
	if
	  Term.literal_equal context_literal Term.minus_v
	  ||
	  Term.literal_equal context_literal Term.plus_v
	then begin
	  regress_one' (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) to_regress
	end
	  

	else begin
	  let literal_info' =
	    try
	      State.get_literal_info state context_literal
	    with
	      | Not_found ->
		  failwith ("Lemma.regress_one: context literal " ^ Term.literal_to_string context_literal)
	  in
	  let explanation =
	    State.get_explanation state [| context_literal |]
	  in
	    match explanation with
	      | [] ->
		  (* a assert at the root chocie point -
		     entailed by the input, just add the clause literal *)
		  regress_one' (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) to_regress
		      
	      | creation_point :: _ when
		  (State.compare_age regress_stop creation_point > 0) ->
		  (* an assert older than the choice points within regression is done -
		     no regression, just add the clause literal *)
		  regress_one' (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) to_regress
		    
	      | _ ->
		  (* a (left) Split - no regression, just add the clause literal  *)
		  if (Array.length literal_info'.State.li_context_literals = 0) then begin
		    regress_one' (index + 1) ((clause_literal, offset) :: regressed) (regressed_size + 1) to_regress
		  end
	      
		  (* the normal case, an actual assert literal to regress *)
		  else begin
		    let new_offset =
		      next_offset := !next_offset + 1;
		      !next_offset
		    in
		    let new_to_regress = {
		      tr_context_literal = context_literal;
		      tr_clause_literal = clause_literal;
		      tr_clause_offset = offset;
		      tr_regress_offset = new_offset;
		      tr_regress_info = literal_info';
		    }
		    in
		      regress_one'
			(index + 1)
			regressed
			regressed_size
			(To_regress.add new_to_regress to_regress)
		  end
	end
    end
  in
    regress_one' 0 regressed regressed_size to_regress



(* regress all context literals to regress *)
let rec regress'
    (state: state)
    (* only regress context literals up to this choice point *)
    (regress_stop: choice_point)
    (* clause literal * offset contained in the regressed literal set *)
    (regressed: (literal * int) list)
    (* size of regressed list, passed along for efficiency *)
    (regressed_size: int)
    (* unification constraints on the regressed literal set *)
    (constraints: ((literal * int) * (literal * int)) list)
    (* branch nodes to regress * offset *)
    (to_regress: To_regress.t)
    (* fresh offsets for the regression *)
    (next_offset: int ref)
    :
    (* the regresset set *)
    ((literal * int) list)
    *
    (* the constraints *)
    (((literal * int) * (literal * int)) list)
    =

  (* nothing more to regress *)
  if To_regress.is_empty to_regress then begin
    regressed, constraints
  end

  (* if we reached the UIP, stop right here *)
  else if uip_reached to_regress then begin
    stop_regression regressed constraints to_regress
  end
  
  (* hard stop of regression *)
  else if regressed_size > Const.max_regression_size then begin
    stop_regression regressed constraints to_regress
  end

  (* add constraint and regress *)
  else begin    
    let next =
      To_regress.max_elt to_regress
    in
    let tail =
      To_regress.remove next to_regress
    in
      
    let literal_info =
      try
	State.get_literal_info state next.tr_context_literal
      with
	| Not_found ->
	    failwith ("Lemma.regress: " ^ Term.literal_to_string next.tr_context_literal)
    in

    (* compute the constraint between the clause literal
       and the context literal to regress *)
    let new_constraints =
      (* the closing clause - start of regression, so no constraint, ignore it *)
      if Term.literal_equal next.tr_clause_literal Term.null_literal then begin
	constraints
      end
	
      else begin
	let new_constraint =
	  (* clause literal paired with the context literal to regress *)
	  (next.tr_clause_literal, next.tr_clause_offset),
	  (* clause literal instantiated to the context literal *)
	  (List.nth literal_info.State.li_clause literal_info.State.li_clause_index, next.tr_regress_offset)
	in
	  new_constraint :: constraints
      end
    in
      
    let regressed', regressed_size', to_regress' =
      regress_one state regress_stop regressed regressed_size tail next_offset next
    in
      regress' state regress_stop regressed' regressed_size' new_constraints to_regress' next_offset
  end




(* wrapper for regress'.
   regresses all context literals to regress. *)
let regress
    (state: state)
    (* only regress context literals up to this choice point *)
    (regress_stop: choice_point)
    (* clause literal * offset contained in the regressed literal set *)
    (to_regress: To_regress.t)
    :
    (* the regresset set *)
    ((literal * int) list)
    *
    (* the constraints *)
    (((literal * int) * (literal * int)) list)
    =
  regress'
    state
    regress_stop
    []
    0
    []
    to_regress
    (ref 1)











      
(*** simplification ***)


(* this is basically a condensing algorithm for the clause
   represented by the list of regressed literals,
   with a specific prefilter due to the nature of their creation. *)
let simplify (state: state) (literals: literal list) : literal list =

  (* first partition the literals into equivalence classes
     based on variable containment.
     I.e. all literals sharing a variable are in the same class. *)

  (* variable to representative mapping *)
  let var_rep : (int, var) Hashtbl.t =
    Hashtbl.create 64
  in
  let literal_rep : Term.var LiteralTable.t =
    LiteralTable.create 128
  in
  let ground_literals : literal list ref =
    ref []
  in

  (* get representative of a variable.
     initially each variable is its own representative. *)
  let rec get_rep (var: var) : var =
    try
      let rep =
	Hashtbl.find var_rep (Var.id_of_var var)
      in
	(* no path compression, so do it recursively *)
	get_rep rep
    with
      | Not_found ->
	  var
  in


  (* merge equivalence classes of variables,
     i.e. let them point to the same representative *)
  let merge_var_classes vars =
    match vars with
      | [] ->
	  failwith "Lemma.merge_var_classes"

      | var :: tail ->
	  let rep =
	    get_rep var
	  in
	    List.iter
	      (fun var ->
		 let cur_rep =
		   get_rep var
		 in
		   if not (Var.equal cur_rep rep) then
		     Hashtbl.add var_rep (Var.id_of_var cur_rep) rep
	      )
	      tail
  in

    (* put each term in its equivalence class *)
    List.iter
      (fun literal ->
	 (* ignore duplicate literals *)
	 if
	   List.exists (Term.literal_equal literal) !ground_literals
	   ||
	   LiteralTable.mem literal_rep literal
	 then begin
	   ()
	 end

	 else begin
	   let vars =
	     Term.vars_of_literal literal
	   in
	     match vars with
	       | [] ->
		   (* as all ground literals remain in the condensed clause,
		      we can stop early if the max. lemma size is exceeded. *)
		   if List.length !ground_literals >= Const.max_lemma_size then begin
		     raise MAX_REGRESSED
		   end;
		   
		   ground_literals := literal :: !ground_literals
		     
	       | var :: _ ->
		   merge_var_classes vars;
		   LiteralTable.add literal_rep literal var
	 end
      )
      literals;

    (* mapping from representative to all literals in its equivalence class *)
    let eq_classes : (int, literal list ref) Hashtbl.t =
      Hashtbl.create 128
    in
      (* move each literal to the final representative *)
      LiteralTable.iter
	(fun literal rep ->
	   let current =
	     try
	       Hashtbl.find eq_classes (Var.id_of_var (get_rep rep))
	     with
	       | Not_found ->
		   let current =
		     ref []
		   in
		     Hashtbl.add eq_classes (Var.id_of_var (get_rep rep)) current;
		     current
	   in
	     (* over-eager filter of too big lemmas.
		misses too much, but prevent waste of computation. *)
	     if List.length !current >= 2 * Const.max_lemma_size then begin
	       raise MAX_REGRESSED
	     end;

	     current := literal :: !current;
	)
	literal_rep;

      
      (* simplify each equivalence class, i.e. each clause, and drop redundant clauses. *)
      let _, clauses =
	Hashtbl.fold
	  (fun _ clause ((length, clauses) as acc) ->
	     let clause =
	       Unification.condense !clause
	     in
	       (* ignore a variant of another clause *)
	       if
		 List.exists
		   (fun clause' ->
		      Term.are_clauses_weak_variants clause clause'
		   )
		   clauses
	       then
		 acc
		   
	       else begin
		 let new_length =
		   length + List.length clause
		 in
		   (* over-eager filter of too big lemmas.
		      misses too much, but prevent waste of computation. *)
		   if new_length > 2 * Const.max_lemma_size then begin
		     raise MAX_REGRESSED
		   end;
		   
		   (new_length, clause :: clauses)
	       end
	  )
	  eq_classes
	  (0, [])
      in

      (* merge all clauses into one big clause *)
      let clause =
	List.fold_left
	  (fun acc clause ->
	     List.fold_left
	       (fun acc literal ->
		  literal :: acc
	       )
	       acc
	       clause
	  )
	  !ground_literals
	  clauses
      in

      (* and finally condense it *)
      let condensed =
	Unification.condense clause
      in
	condensed





(*** main function ***)


let get_lemma (state: state) (clause: clause) (conflict_literals: literal array)  : clause option =

  let closing_explanation =
    State.get_explanation state conflict_literals
  in

  let regress_stop =
    match closing_explanation with
      | [] ->
	  failwith "Lemma.get_lemma: regression stop"
	  
      | head :: _ ->
	  head
  in

  (* fake data to get the regression with the closing clause started *)
  let closing_info = {
    State.li_literal = Term.null_literal;
    State.li_choice_point = State.active_choice_point state;
    State.li_clause = clause;
    State.li_clause_index = -1;
    State.li_context_literals = conflict_literals;
    State.li_explanation = Some closing_explanation;
    State.li_clause_utility = (fun () -> ());
  }
  in
  let regression_start = {
    tr_clause_literal = Term.null_literal;
    tr_clause_offset = -1;
    tr_context_literal = Term.null_literal;
    tr_regress_offset = 0;
    tr_regress_info = closing_info;
  }
  in

  (* do the regression *)
  let regressed_literals, constraints =
    regress
      state
      regress_stop
      (To_regress.add regression_start To_regress.empty)
  in

  (* create the conflict set by applying the constraints to the regressed literals *)
  let regressed =
    Subst.apply_to_literals
      (Unification.unify_constraints constraints)
      regressed_literals
  in

    try
      let simplified =
	simplify state regressed
      in
	
	(* lemma too long *)
	if List.length simplified > Const.max_lemma_size then begin
	  None
	end
	  
	else begin
	  (* a lemma is universal *)
	  let universal =
	    Term.request_universal_clause (Term.sort_clause simplified)
	  in
	    (* normalize if its a unit lemma.
	       literals of clauses are automatically normalized when
	       chosen for assert or split by application of the context unifier,
	       unless they are unit, as then no context unifier is applied.
	    *)
	  let universal =
	    match universal with
	      | [ literal ] ->
		  [ Subst.apply_to_literal Subst.empty literal 0 ]
		    
	      | _ ->
		  universal
	  in
	    Some universal;
	end
	  
    with
      | MAX_REGRESSED ->
	  (* lemma computation to expensive, aborted *)
	  None


