(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2005
              The University of Iowa

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type config = Config.config
type state = State.state
type literal = Term.literal
type clause = Term.clause
type choice_point = State.choice_point
type explanation = State.explanation
type branch_node = State.literal_info
type raw_context_unifier = Context_unifier.raw_context_unifier


exception UNSATISFIABLE = Context_unifier.UNSATISFIABLE



(*** scoring ***)




let rec score' (state: state) (scored: clause list) (to_score: literal list) : unit =
  match to_score with
    | [] ->
	()

    | context_literal :: tail ->
	let literal_info =
	  try
	    State.get_literal_info state context_literal
	  with
	    | Not_found ->
		failwith ("State_backtrack: score' " ^ Term.literal_to_string context_literal)
	in

	  (* score each clause only once *)
	  if List.memq literal_info.State.li_clause scored then begin
	    score' state scored tail
	  end

	  else begin
	    (* score clause *)
	    literal_info.State.li_clause_utility ();

            (* add the context literals to be scored as well.
	       for a split no context literals are given, so scoring is stopped automatically.  *)
            let rec score_add' index to_score =
	      if index >= Array.length literal_info.State.li_context_literals then
		to_score

	      else
		let context_literal =
		  literal_info.State.li_context_literals.(index)
		in
		  (* ignore pseudo context literals *)
		  if
		    Term.literal_equal context_literal Term.assert_literal
		    ||
		    Term.literal_equal context_literal Term.minus_v
		    ||
		    Term.literal_equal context_literal Term.plus_v
		  then
		    score_add' (index + 1) to_score

		  else
		    score_add' (index + 1) (context_literal :: to_score)
	    in
	    let to_score' =
	      score_add' 0 to_score
	    in
	      score' state (literal_info.State.li_clause :: scored) to_score'

	  end
  

let score (state: state) (closing_context_unifier: raw_context_unifier) : unit =
  let clause =
    closing_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
  in
  let context_literals =
    Context_unifier.get_context_literals
      closing_context_unifier.Context_unifier.rcu_context_partners
  in

    (* score closing clause *)
    closing_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value :=
      !(closing_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value) + 1;
	
    (* score all other involved clauses *)
    score' state [clause] (Array.to_list context_literals)



(*** backtracking ***)


(* invalidates and removes from the derivation branch
   all choice points older than and including the choice point to retract. *)
let invalidate (state: state) (retract: choice_point) : unit =

  let rec invalidate' choice_points =
    match choice_points with
      | head :: tail ->
	  if State.compare_age retract head < 0 then begin
	    State.invalidate_choice_point head;
	    invalidate' tail
	  end

	  else if State.compare_age retract head = 0 then begin
	    State.invalidate_choice_point head;
	    State.set_branch state tail;
	  end

	  else begin
	    failwith "State_backtrack.invalidate 1"
	  end
	      
      | [] ->
	  failwith "State_backtrack.invalidate 2"
  in
    invalidate' (State.get_branch state)






(* backtrack state based on the closing context unifier *)
let backtrack (state: state) (config: config) (closing_context_unifier: raw_context_unifier) :
    choice_point * explanation * clause option
    =
  let clause =
    closing_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
  in
  let context_literals =
    Context_unifier.get_context_literals
      closing_context_unifier.Context_unifier.rcu_context_partners
  in

  let closing_explanation =
    match Config.backtracking config with
      | Flags.BT_Naive ->
	  State.get_branch state
	  
      | Flags.BT_Backjumping ->
	  State.get_explanation state context_literals
  in

  let retract, retract_next, right_explanation =
    match closing_explanation with
      | [] ->
	  raise UNSATISFIABLE
	  
      | root :: [] ->
	  if
	    State.choice_point_equal (State. root_choice_point state) root
	    ||
	    State.choice_point_equal State.valid_choice_point root
	  then
	    raise UNSATISFIABLE
	  else
	    root, (State.root_choice_point state), []

      | retract :: retract_next :: tail ->
	  retract, retract_next, (retract_next :: tail)
  in

  let backjump =
    (* how much do we actually want to backtrack?
       we could go right to the backjump target, like SAT solvers do.
       but in our case that mostly leads to the recomputation
       of most of the jumped over derivation, and thus a performance loss.
       therefore, just retract the most recent responsible left split. *)
    retract;
    (*
    let rec find_backjump branch =
      match branch with
       | [] ->
	   failwith "State_backtrack.find_backjump 0"
	     
       | _ :: [] ->
	   failwith "State_backtrack.find_backjump 1"
	     
       | x :: y :: tail ->
	   if State.choice_point_equal y retract_next then
	     x
	   else
	     find_backjump (y :: tail)
    in
      find_backjump (State.get_branch state)
    *)
  in

  (* lemma must be computed before the used dependency information is backtracked *)
  let lemma =
    if (Config.lemma config) then
      Lemma.get_lemma state clause context_literals
    else
      None
  in

    (* scoring must be done before the used dependency information is backtracked *)
    score state closing_context_unifier;

    invalidate state backjump;

    (* remove invalidated context literals. *)
    State.backtrack_literal_info state;

    retract, right_explanation, lemma


let backtrack_incomplete (state: state) (retract: choice_point) : unit =
  if
    State.choice_point_equal (State.root_choice_point state) retract
    ||
    State.choice_point_equal State.valid_choice_point retract
  then begin
    raise UNSATISFIABLE
  end;

  invalidate state retract;

  (* remove invalidated context literals. *)
  State.backtrack_literal_info state
