(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** Types ***)

type var = Var.var
type term = Term.term
type literal = Term.literal
type clause = Term.clause
type binding = Subst.binding
type subst = Subst.subst
type subst_var = Subst.var
type packed_subst = Subst.packed_subst

exception UNIFICATION_FAIL



(* Like in Subst it pays off performance wise
   to precompile p_preserving and i_preserving in specialized modules *)
module type Unification = 
sig
  val match_terms: term -> int -> term -> int -> subst
  val unify_terms: subst -> term -> int -> term -> int -> subst
  val match_substs: int -> subst -> subst -> subst
  val unify_substs: subst -> subst -> subst
  val unify_packed_substs: subst -> packed_subst -> subst
end


module type Options =
sig
  val p_preserving: bool
  val i_preserving: bool
end


(* functor *)
module Make (Options: Options) : Unification =
struct

  (*** Matching ***)
  let subst_set subst var term =
    if Options.p_preserving && not Options.i_preserving then
      Subst.PPreserving.set subst var term
	
    else if not Options.p_preserving && Options.i_preserving then
      Subst.IPreserving.set subst var term
	
    else if Options.p_preserving && Options.i_preserving then
      Subst.PIPreserving.set subst var term

    else (* Options.p_preserving && Options.i_preserving then *)
      Subst.Preserving.set subst var term

  let subst_set_ subst var offset1 term2 offset2 =
    if Options.p_preserving && not Options.i_preserving then
      Subst.PPreserving.set_ subst var offset1 term2 offset2
	
    else if not Options.p_preserving && Options.i_preserving then
      Subst.IPreserving.set_ subst var offset1 term2 offset2

    else if Options.p_preserving && Options.i_preserving then
      Subst.PIPreserving.set_ subst var offset1 term2 offset2

    else (* Options.p_preserving && Options.i_preserving then *)
      Subst.Preserving.set_ subst var offset1 term2 offset2
	

  let rec do_match_terms (subst: subst) (term1: term) (offset1: int) (term2: term) (offset2: int) : subst =
    
    if (offset1 == offset2) && (Term.term_equal term1 term2) then
      subst
    else
      match term1, term2 with
	| Term.Var var, _ ->
	    begin
	      match Subst.get_ subst var offset1 with
		| Some bound_term ->
		    do_match_terms
		      subst
		      bound_term.Subst.st_term
		      bound_term.Subst.st_offset
		      term2
		      offset2
		      
		| None ->
		    (* term1 is part of the term to match and may thus not be bound. *)
		    if offset1 == offset2 then
		      raise UNIFICATION_FAIL
			
		    (* var can be bound if p/i preserving don't apply *)
		    else subst_set_ subst var offset1 term2 offset2
	  end
	      
      | Term.Const symbol1, Term.Const symbol2 when 
	  Symbol.equal symbol1 symbol2 ->
	  
	  subst
	      
      | Term.Func (symbol1, terms1, _), Term.Func (symbol2, terms2, _) when
	  Symbol.equal symbol1 symbol2 ->
	    Tools.array_fold2
	      (fun subst term1 term2 ->
		 do_match_terms
		   subst term1 offset1 term2 offset2
	      )
	      subst
	      terms1
	      terms2
	      
      | _ -> 
	  raise UNIFICATION_FAIL





  let match_terms (term1: term) (offset1: int) (term2: term) (offset2: int) : subst =
    try
      let subst =
	do_match_terms
	  Subst.empty term1 offset1 term2 offset2
      in 
	subst
    with
      | Subst.SET_FAIL ->
	  raise UNIFICATION_FAIL







  (*** Unification ***)
	  
	  
  let rec do_unify_terms (subst: subst)
      (term1: term) (offset1: int) (term2: term) (offset2: int) : subst =

    if (offset1 == offset2) && (Term.term_equal term1 term2) then 
      subst
    else
      begin
	match term1, term2 with
	  | Term.Var var1, Term.Var var2 ->
	      begin
		(* dereference bound variables *)
		match Subst.get_ subst var1 offset1, Subst.get_ subst var2 offset2 with
		  | Some bound_term1, Some bound_term2 ->
		      do_unify_terms
			subst
			bound_term1.Subst.st_term
			bound_term1.Subst.st_offset
			bound_term2.Subst.st_term
			bound_term2.Subst.st_offset
		      
		  | Some bound_term1, None ->
		      do_unify_terms
			subst
			bound_term1.Subst.st_term
			bound_term1.Subst.st_offset
			term2
			offset2
			
		  | None, Some bound_term2 ->
		      do_unify_terms
			subst
			term1
			offset1
			bound_term2.Subst.st_term
			bound_term2.Subst.st_offset
			
		  | None, None ->
		      (* if a variable is an indicator bind the other variable *)
		      if
			Options.i_preserving
			&&
			Var.is_indicator var1
			&&
			not (Var.is_indicator var2)
		      then
			subst_set_ subst var2 offset2 term1 offset1

		    (* if a variable is an indicator bind the other variable *)
		    else if
		      Options.i_preserving
		      &&
		      Var.is_indicator var2
		      &&
		      not (Var.is_indicator var1)
		    then
		      subst_set_ subst var1 offset1 term2 offset2

		    (* try to bind a universal variable *)
		    else if Var.is_universal var1 then
		      subst_set_ subst var1 offset1 term2 offset2
			
		    else (*if Var.is_universal var2 then*)
		      subst_set_ subst var2 offset2 term1 offset1
	      end
		
	  | Term.Var var1, _ ->
	      begin
		(* dereference bound variable *)
		match Subst.get_ subst var1 offset1 with
		  | Some bound_term1 ->
		      do_unify_terms
			subst
			bound_term1.Subst.st_term
			bound_term1.Subst.st_offset
			term2
			offset2
		      
		  | None ->
		      subst_set_ subst var1 offset1 term2 offset2
	      end
		
	  | _, Term.Var var2 ->
	      begin
		(* dereference bound variable *)
		match Subst.get_ subst var2 offset2 with
		  | Some bound_term2 ->
		      do_unify_terms
			subst
			term1
			offset1
			bound_term2.Subst.st_term
			bound_term2.Subst.st_offset
		      
		  | None ->		    
		      subst_set_ subst var2 offset2 term1 offset1
	      end
		
	  | Term.Const symbol1, Term.Const symbol2 when
	      Symbol.equal symbol1 symbol2 ->
	      
	      subst
		
		
	  | Term.Func (symbol1, terms1, _), Term.Func (symbol2, terms2, _) when
	      Symbol.equal symbol1 symbol2 ->
	      
	      Tools.array_fold2
		(fun subst term1 term2 ->
		   do_unify_terms
		     subst
		     term1
		     offset1
		     term2
		     offset2
		)
		subst
		terms1
		terms2
		
	  | _ ->
	      raise UNIFICATION_FAIL
      end
    


  let unify_terms
      (subst: subst) (term1: term) (offset1: int) (term2: term) (offset2: int) : Subst.subst =
    try  
      do_unify_terms subst term1 offset1 term2 offset2
    with
      | Subst.SET_FAIL ->
	  raise UNIFICATION_FAIL







  (*** Merging ***)
	    
  (* does subst contain binding reversed? *)	
  let is_included_reversed (subst: subst) (binding: binding) : bool =
    match binding.Subst.sb_term.Subst.st_term with
      | Term.Var term_var ->
	  Subst.exists
	    (fun current ->
	       (current.Subst.sb_var.Subst.sv_offset == binding.Subst.sb_term.Subst.st_offset)
	       &&
	       (Var.equal current.Subst.sb_var.Subst.sv_var term_var)
	       &&  
	       (current.Subst.sb_term.Subst.st_offset = binding.Subst.sb_var.Subst.sv_offset)
	       &&
		 (match current.Subst.sb_term.Subst.st_term with
		    | Term.Var binding_bound_var when
			Var.equal binding_bound_var binding.Subst.sb_var.Subst.sv_var ->
		     	true
			  
		    | _ ->
			false
		 )
	    )
	    subst
	    
      | _ ->
	  false
          


  let do_unify_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
      (acc: subst) (binding: binding) : subst =
    
    let var, term =
      binding.Subst.sb_var, binding.Subst.sb_term
    in
      match Subst.get acc var with
	| None ->
	    if is_included_reversed acc binding then
	      acc
	    else
	      (* var is unbound till now, so just keep its binding *)
	      subst_set acc var term
	      
	| Some acc_term ->
	    (* var is already bound, so unify the existing bindings *)
	    do_unify_terms
	      acc
	      term.Subst.st_term term.Subst.st_offset
	      acc_term.Subst.st_term acc_term.Subst.st_offset


  let unify_substs (first: subst) (second: subst) : subst =
    if Subst.is_empty first then
      second
    else if Subst.is_empty second then
      first
    else
      begin
        try
	  Subst.fold
	    (fun acc binding ->
	       do_unify_substs acc binding
	    )
	    first
	    second
        with
          | Subst.SET_FAIL ->
	      raise UNIFICATION_FAIL
      end
	    
	    
  let unify_packed_substs (first: subst) (second: packed_subst) : subst =
    if Subst.is_empty first then
      Subst.unpack_subst second
            
    else
      begin
        try
	  Array.fold_left
	    (fun acc binding ->
	       do_unify_substs acc binding
	    )
	    first
	    second
        with
          | Subst.SET_FAIL ->
	      raise UNIFICATION_FAIL
      end






  let do_match_substs (more_general_offset: int) (acc: subst) (subst: subst) : subst =
    Subst.fold
      (fun acc binding ->
	 let var, term =
	   binding.Subst.sb_var, binding.Subst.sb_term
	 in
	   match Subst.get acc var with
	     | None ->
		 (* var is unbound till now, so just keep its binding *)
		 subst_set acc var term
	       
	     | Some acc_term ->
		 (* var is already bound, so unify the existing bindings *)
		 if
		   term.Subst.st_offset = more_general_offset
		   &&
		   acc_term.Subst.st_offset = more_general_offset
		 then begin
		   (* unify *)
		   do_unify_terms
		     acc
		     term.Subst.st_term term.Subst.st_offset
		     acc_term.Subst.st_term acc_term.Subst.st_offset
		 end

		 else if term.Subst.st_offset = more_general_offset then begin
		   (* match term to acc *)
		   do_match_terms
		     acc
		     term.Subst.st_term term.Subst.st_offset
		     acc_term.Subst.st_term acc_term.Subst.st_offset
		 end
		   
		 else if acc_term.Subst.st_offset = more_general_offset then begin
		   (* match acc to term *)
		   do_match_terms
		     acc
		     acc_term.Subst.st_term acc_term.Subst.st_offset
		     term.Subst.st_term term.Subst.st_offset
		 end
		   
		 else begin
		   (* both terms part of the term to match -
		      no variables can be bound *)
		   if Term.term_equal term.Subst.st_term acc_term.Subst.st_term then
		     acc
		   else
		     raise UNIFICATION_FAIL
		 end
		   
      )
      acc
      subst


  let match_substs (more_general_offset: int) (first: subst) (second: subst) : subst =
    if Subst.is_empty first then
      second

    else if Subst.is_empty second then
      first

    else
      begin
        try
          do_match_substs
	    more_general_offset first second
        with
          | Subst.SET_FAIL ->
	      raise UNIFICATION_FAIL
      end
end

  
module UnificationPPreserving =
  Make (
    struct
      let p_preserving = true
      let i_preserving = false
    end
  )

module UnificationIPreserving =
  Make (
    struct
      let p_preserving = false
      let i_preserving = true
    end
  )

module UnificationPIPreserving =
  Make (
    struct
      let p_preserving = true
      let i_preserving = true
    end
  )

module Unification_Preserving =
  Make (
    struct
      let p_preserving = false
      let i_preserving = false
    end
  )




(* the general wrapper functions for the spezialiced modules *)
let match_terms ?(p_preserving: bool = false) ?(i_preserving:bool = false)
  (term1: term) (offset1: int) (term2: term) (offset2: int) : subst =

  if p_preserving && not i_preserving then
    UnificationPPreserving.match_terms term1 offset1 term2 offset2

  else if not p_preserving && i_preserving then
    UnificationIPreserving.match_terms term1 offset1 term2 offset2

  else if p_preserving && i_preserving then
    UnificationPIPreserving.match_terms term1 offset1 term2 offset2

  else (* not p_preserving && not i_preserving then *)
    Unification_Preserving.match_terms term1 offset1 term2 offset2

let match_literals ?(p_preserving: bool = false) ?(i_preserving:bool = false)
  (literal1: literal) (offset1: int) (literal2: literal) (offset2: int) : subst =

  if (literal1.Term.sign = literal2.Term.sign) then
    match_terms
      ~p_preserving:p_preserving ~i_preserving:i_preserving
      literal1.Term.term offset1 literal2.Term.term offset2
  else
    raise UNIFICATION_FAIL


let is_term_generalization ?(p_preserving: bool = false) ?(i_preserving:bool = false)
    (general: term) (instance: term) : bool =
  try
    ignore (
      match_terms ~p_preserving:p_preserving ~i_preserving:i_preserving
        general 0 instance 1 : subst
    );
    true
  with
    | UNIFICATION_FAIL ->
        false

let is_term_instance ?(p_preserving: bool = false) ?(i_preserving:bool = false)
    (instance: term) (general: term) : bool =
  is_term_generalization ~p_preserving:p_preserving ~i_preserving:i_preserving general instance
	  


let unify_terms ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (term1: term) (offset1: int) (term2: term) (offset2: int) : Subst.subst =

  if p_preserving && not i_preserving then
    UnificationPPreserving.unify_terms Subst.empty term1 offset1 term2 offset2

  else if not p_preserving && i_preserving then
    UnificationIPreserving.unify_terms Subst.empty term1 offset1 term2 offset2

  else if p_preserving && i_preserving then
    UnificationPIPreserving.unify_terms Subst.empty term1 offset1 term2 offset2

  else (* not p_preserving && not i_preserving then *)
    Unification_Preserving.unify_terms Subst.empty term1 offset1 term2 offset2


let unify_literals ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (literal1: literal) (offset1: int) (literal2: literal) (offset2: int) : Subst.subst =

  if literal1.Term.sign = literal2.Term.sign then
    unify_terms ~p_preserving:p_preserving ~i_preserving:i_preserving
      literal1.Term.term offset1 literal2.Term.term offset2
  else
    raise UNIFICATION_FAIL

let are_terms_unifiable ?(p_preserving: bool = false) ?(i_preserving:bool = false)
    (x: term) (y: term) : bool =
  try
    ignore (
      unify_terms ~p_preserving:p_preserving ~i_preserving:i_preserving
        x 0 y 1 : subst
    );
    true
  with
    | UNIFICATION_FAIL ->
        false




let unify_constraints (constraints : ((literal * int) * (literal * int)) list) : subst =
  List.fold_left
    (fun subst ((literal1, offset1), (literal2, offset2)) ->
       Unification_Preserving.unify_terms
	 subst
	 literal1.Term.term offset1
	 literal2.Term.term offset2
      )
    Subst.empty
    constraints







let unify_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (first: subst) (second: subst) : subst =

  if p_preserving && not i_preserving then
    UnificationPPreserving.unify_substs first second

  else if not p_preserving && i_preserving then
    UnificationIPreserving.unify_substs first second

  else if p_preserving && i_preserving then
    UnificationPIPreserving.unify_substs first second

  else (* p_preserving && i_preserving then *)
    Unification_Preserving.unify_substs first second

let match_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (more_general_offset: int) (first: subst) (second: subst) : subst =

  if p_preserving && not i_preserving then
    UnificationPPreserving.match_substs more_general_offset first second

  else if not p_preserving && i_preserving then
    UnificationIPreserving.match_substs more_general_offset first second

  else if p_preserving && i_preserving then
    UnificationPIPreserving.match_substs more_general_offset first second

  else (* p_preserving && i_preserving then *)
    Unification_Preserving.match_substs more_general_offset first second


let unify_packed_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (first: subst) (second: packed_subst) : subst =

  if p_preserving && not i_preserving then
    UnificationPPreserving.unify_packed_substs first second

  else if not p_preserving && i_preserving then
    UnificationIPreserving.unify_packed_substs first second

  else if p_preserving && i_preserving then
    UnificationPIPreserving.unify_packed_substs first second

  else (* p_preserving && i_preserving then *)
    Unification_Preserving.unify_packed_substs first second

let merge_substs (first: subst) (second: subst) : subst =
  if Subst.is_empty first then
    second
      
  else if Subst.is_empty second then
    first
      
  else
    begin
      Subst.fold
	(fun acc binding ->
	   let var, term =
	     binding.Subst.sb_var, binding.Subst.sb_term
	   in
	     match Subst.get acc var with
	       | None ->
		   (* var is unbound till now, so just keep its binding *)
		   Subst.append binding acc
		     
	       | Some acc_term ->
		   (* this binding does already exists *)
		   if Subst.term_equal term acc_term then begin
		     acc
		   end

		   (* same variable bound to different terms *)
		   else
		     raise UNIFICATION_FAIL
			 
	)
	first
	second
    end








(*** subsumption ***)  





let subsumes (_subsuming: clause) (_subsumed: clause) : bool =

  let rec subsumes' (subsuming: clause) (subst: subst) : unit =
    match subsuming with
      | [] ->
	  raise Exit

      | general :: tail ->
	  List.iter
	    (fun instance ->
	       try
		 let new_subst =
		   match_literals
		     general 0
		     instance 1
		 in
		 let merged_subst =
		   merge_substs
		     subst
		     new_subst
		 in
		   (* ok, try to extend this potential match *)
		   subsumes' tail merged_subst
	       with
		 | UNIFICATION_FAIL ->
		     (* nope, try next literal *)
		     ()
	    )
	    _subsumed
  in
    try
      subsumes' _subsuming Subst.empty;
      false
    with
      | Exit ->
	  true



(* algorithm taken from
   Gottlob, Fermueller. Removing redundancy from a clause.

   iteratively check for each clause literal:
   if the current condensed clause C (starting with the original clause)
   is subsumed by C' = C \ {L}, then C' is the new condensed clause.
*)
let condense (clause: clause) : clause =

  let rec condense' (condensed: clause) (to_process: literal list) : clause =
    match to_process with
      | [] ->
	  (* all clause literals checked, so this is the condensed literal *)
	  condensed

      | head :: tail ->
	  (* a ground literal must be part of the condensed clause -
	     assuming no literal exists more than once in a clause. *)
	  if Term_attributes.is_term_ground head.Term.term then begin
	    condense' condensed tail
	  end
	    
	  (* cheack precheck:
	     head must be more general than another clause literal. *)
	  else if
	    List.for_all
	      (fun x ->
		 (Term.literal_equal head x)
		 ||
		 begin
		   try
		     ignore (match_literals head 0 x 1 : Subst.subst);
		     false
		   with
		     | UNIFICATION_FAIL ->
			 true
		 end
	      )
	      condensed
	  then begin
	    condense' condensed tail
	  end


	  else begin
	    (* the potentially condensed clause *)
	    let condensed' =
	      Tools.list_remove_first
		(Term.literal_equal head)
		condensed
	    in
	      if subsumes condensed condensed' then
		(* yes, condensed *)
		condense' condensed' tail
	      else
		(* no, keep previous clause *)
		condense' condensed tail
	  end
  in
  let x =
    condense' clause clause
  in
    (*
    if List.length x < List.length clause then begin
      Statistic.inc_global_debug ();
      print_endline ("CLAUSE   :\n" ^ Term.clause_to_string clause);
      print_endline ("CONDENSED:\n" ^ Term.clause_to_string x);
    end;
    *)
    x
