(* $Id: context.mli,v 1.19 2004/12/28 14:06:46 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** context with operations

    Internally applies the {e Compact} rule of the model evolution calculus.

    As only order preserving (chronological, backjumping) backtracking is used,
    the context can be seen as a stack of literals,
    backtracking always removes the last elements that were added.
*)

(** {6 Types} *)

type config = Config.config
type statistic = Statistic.statistic
type counter = Counter.counter
type var = Var.var
type term = Term.term
type literal = Term.literal
type clause = Term.clause
type choice_point = State.choice_point
type state = State.state


(** a context literal with some information *)
type element = {
  el_id: int;
  (** unique id per context literal. *)

  el_literal: literal;
  (** the context literal. *)

  el_choice_point: choice_point;
  (** the choice point in which the context literal was added. *)

  el_pars: var list;
  (** the parameters of the context literal. *)

  el_generation: int;
  (** depth of the tree of context literals
      needed to compute this context literal. *)

  el_implied_by: int;
  (** if <> -1, then this context literal is implied by other context literals,
      i.e. it as been added by an Assert or Unit Split.
      then, this is the id of the most recently added implying context literal.
      
      this can be exploited,
      to permanently drop candidate literals
      which are blocked by this literal (e.g. subsumed).
      if they are computed by a context unifier using a context literal
      with id >= el_implied_by,
      then they will be blocked for their whole lifetime. *)

  mutable el_compacted: int;
  (** the earliest choice point in which a literal was added to the context
      which compacts this literal. *)
}


type context



(** {6 Functions} *)




(** {2 Initialization} *)

(** empties the context and fixes the indexing method based. *)
val create: config -> statistic -> state -> context




(** {2 Constants} *)

(** for initialization usage, invalid choice point. *)
val null_element: element

(** pseudo assert literal for use as the assert gap in a context unifier,
    contains {!Term.assert_literal}, always valid choice point. *)
val assert_element: element

(** pseudo plus_v literal for use in a context unifier,
    contains {!Term.plus_v_literal}, always valid choice point. *)
val plus_v_element: element

(** pseudo minus_v literal for use in a context unifier,
    contains {!Term.minus_v_literal}, always valid choice point. *)
val minus_v_element: element


(** {2 Access} *)

(** [add context literal implied_by generation]
    adds a new [literal] to the [context],
    for [implied_by] and [generation] see {!Context.element}.
*)
val add: context -> literal -> int -> int -> element

(** the same element? *)
val is_equal: element -> element -> bool



(** {2 Iteration} *)

val iter: context -> (element -> unit) -> unit

val fold: context -> ('a -> element -> 'a) -> 'a -> 'a



(** {2 Checks} *)

(** [check_contradictory context literal]
    checks if [literal] is contradictory with the [context].
    Returns a contradictory context element, if one exists. *)
val check_contradictory: context -> literal -> element option

(** [check_subsumed context literal]
    checks if [literal] is subsumed by the [context].
    Returns a subsuming context element, if one exists. *)
val check_subsumed: context -> literal -> element option

(** [check_productive context producer produced]
    checks if [producer] is producing the negation of [produced] in the [context]
    (that is [producer] is the context literal,
    and [produced] the remainder literal).
    Returns a conflicting context element, if one exists. *)
val check_productive: context -> literal -> literal -> element option

(** [is_satisfied context literal]
    Returns true, if the literal is satisfied by the [context]. 
    
    Note: this is not complete but merely a prefilter with linear complexity,
    as the full test is too expensive. *)
val is_satisfied: context -> literal -> bool



(** {2 Properties} *)

(** [is_present element element]
    checks if either
    - the element has an id lower or equal to [context_literal_id]
    - or the element is implied by the context
    up to and including the context literal with the id [context_literal_id].
*)
val is_present: element -> element -> bool



(** {2 Representation} *)

val element_to_string: element -> string


(** [from_file file_name]
  creates a context from a file.
  The file should contain one literal per line, readable by {!Read_tme}.

  @raise Sys_error on file access failure.

  @raise Failure if the file contains invalid input. *)
val from_file: string ->
  config -> statistic -> state -> context


(** print the (not compacted) context literals, one literal per line.
    Compacted literals are only shown when [show_compacted] is true
    (by default it is false). *)
val print_context: ?show_compacted:bool -> context -> out_channel -> unit

(** prototypical implementation of the DIG representation.
    Works currently only for {e -v} *)
val print_DIG: context -> out_channel -> clause list -> unit


(** string representation of the maximimal context size
    during the derivation (ala {!Print.print_label}). *)
val print_max_size: context -> unit



(** {2 Backtracking} *)

(** removes context elements added in invalid choice points,
    and undoes the corresponding compacting of the context.
    [choice_point] is the oldest choice point retracted
    since the last backtracking.
*)
val backtrack: context -> unit

(* returns-10 if no element exists, its id starting from 1 otherwise *)
val get_max_id: context -> int
(*val get_max_prev_id: context -> int
val next_el: context -> int -> int*)
